//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// cpu6502.cpp: Implementation of 6502 CPU 
//////////////////////////////////////////////////////////////////////

#include "Cpu6502.h"

//////////////////////////////////////////////////////////////////////
// Procedure: cpu6502 Constructor
//////////////////////////////////////////////////////////////////////
// Description: The constructor is only in charge of making sure 
// memory is a NULL pointer and building a look-up table for instruction
// addressing modes.  For the actual initialization, look to the
// Init procedure
//////////////////////////////////////////////////////////////////////
cpu6502::cpu6502()
{	
	// Make sure memory pointer is null
	memory = NULL;
	
	// Build instruction/addressing mode table
	instruction_table[ADC_IMMEDIATE] = IMMEDIATE;
	instruction_table[ADC_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[ADC_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[ADC_ABSOLUTE] = cABSOLUTE;
	instruction_table[ADC_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[ADC_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[ADC_INDIRECT_X] = INDIRECT_X;
	instruction_table[ADC_INDIRECT_Y] = INDIRECT_Y;

	instruction_table[AND_IMMEDIATE] = IMMEDIATE;
	instruction_table[AND_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[AND_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[AND_ABSOLUTE] = cABSOLUTE;
	instruction_table[AND_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[AND_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[AND_INDIRECT_X] = INDIRECT_X;
	instruction_table[AND_INDIRECT_Y] = INDIRECT_Y;
		
	instruction_table[ASL_ACCUMULATOR] = ACCUMULATOR;
	instruction_table[ASL_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[ASL_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[ASL_ABSOLUTE] = cABSOLUTE;
	instruction_table[ASL_ABSOLUTE_X] = ABSOLUTE_X;
	
	instruction_table[BCC_RELATIVE] = cRELATIVE;
	instruction_table[BCS_RELATIVE] = cRELATIVE;
	instruction_table[BEQ_RELATIVE] = cRELATIVE;
	
	instruction_table[BIT_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[BIT_ABSOLUTE] = cABSOLUTE;

	instruction_table[BMI_RELATIVE] = cRELATIVE;
	instruction_table[BNE_RELATIVE] = cRELATIVE;
	instruction_table[BPL_RELATIVE] = cRELATIVE;
	
	instruction_table[BRK_IMPLIED] = IMPLIED;

	instruction_table[BVC_RELATIVE] = cRELATIVE;
	instruction_table[BVS_RELATIVE] = cRELATIVE;

	instruction_table[CLC_IMPLIED] = IMPLIED;
	instruction_table[CLD_IMPLIED] = IMPLIED;
	instruction_table[CLI_IMPLIED] = IMPLIED;
	instruction_table[CLV_IMPLIED] = IMPLIED;
	
	instruction_table[CMP_IMMEDIATE] = IMMEDIATE;
	instruction_table[CMP_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[CMP_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[CMP_ABSOLUTE] = cABSOLUTE;
	instruction_table[CMP_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[CMP_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[CMP_INDIRECT_X] = INDIRECT_X;
	instruction_table[CMP_INDIRECT_Y] = INDIRECT_Y;

	instruction_table[CPX_IMMEDIATE] = IMMEDIATE;
	instruction_table[CPX_ZERO_PAGE] = ZERO_PAGE;	
	instruction_table[CPX_ABSOLUTE] = cABSOLUTE;

	instruction_table[CPY_IMMEDIATE] = IMMEDIATE;
	instruction_table[CPY_ZERO_PAGE] = ZERO_PAGE;	
	instruction_table[CPY_ABSOLUTE] = cABSOLUTE;

	instruction_table[DEC_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[DEC_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[DEC_ABSOLUTE] = cABSOLUTE;
	instruction_table[DEC_ABSOLUTE_X] = ABSOLUTE_X;

	instruction_table[DEX_IMPLIED] = IMPLIED;
	instruction_table[DEY_IMPLIED] = IMPLIED;

	instruction_table[EOR_IMMEDIATE] = IMMEDIATE;
	instruction_table[EOR_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[EOR_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[EOR_ABSOLUTE] = cABSOLUTE;
	instruction_table[EOR_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[EOR_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[EOR_INDIRECT_X] = INDIRECT_X;
	instruction_table[EOR_INDIRECT_Y] = INDIRECT_Y;
	
	instruction_table[INC_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[INC_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[INC_ABSOLUTE] = cABSOLUTE;
	instruction_table[INC_ABSOLUTE_X] = ABSOLUTE_X;
	
	instruction_table[INX_IMPLIED] = IMPLIED;
	instruction_table[INY_IMPLIED] = IMPLIED;

	instruction_table[JMP_ABSOLUTE] = cABSOLUTE;
	instruction_table[JMP_INDIRECT] = INDIRECT;
	
	instruction_table[JSR_ABSOLUTE] = SPECIAL;

	instruction_table[LDA_IMMEDIATE] = IMMEDIATE;
	instruction_table[LDA_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[LDA_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[LDA_ABSOLUTE] = cABSOLUTE;
	instruction_table[LDA_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[LDA_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[LDA_INDIRECT_X] = INDIRECT_X;
	instruction_table[LDA_INDIRECT_Y] = INDIRECT_Y;

	instruction_table[LDX_IMMEDIATE] = IMMEDIATE;
	instruction_table[LDX_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[LDX_ZERO_PAGE_Y] = ZERO_PAGE_Y;
	instruction_table[LDX_ABSOLUTE] = cABSOLUTE;
	instruction_table[LDX_ABSOLUTE_Y] = ABSOLUTE_Y;
	
	instruction_table[LDY_IMMEDIATE] = IMMEDIATE;
	instruction_table[LDY_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[LDY_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[LDY_ABSOLUTE] = cABSOLUTE;
	instruction_table[LDY_ABSOLUTE_X] = ABSOLUTE_X;

	instruction_table[LSR_ACCUMULATOR] = ACCUMULATOR;
	instruction_table[LSR_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[LSR_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[LSR_ABSOLUTE] = cABSOLUTE;
	instruction_table[LSR_ABSOLUTE_X] = ABSOLUTE_X;

	instruction_table[NOP_IMPLIED] = IMPLIED;

	instruction_table[ORA_IMMEDIATE] = IMMEDIATE;
	instruction_table[ORA_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[ORA_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[ORA_ABSOLUTE] = cABSOLUTE;
	instruction_table[ORA_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[ORA_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[ORA_INDIRECT_X] = INDIRECT_X;
	instruction_table[ORA_INDIRECT_Y] = INDIRECT_Y;
		
	instruction_table[PHA_IMPLIED] = IMPLIED;
	instruction_table[PHP_IMPLIED] = IMPLIED;
	instruction_table[PLA_IMPLIED] = IMPLIED;
	instruction_table[PLP_IMPLIED] = IMPLIED;

	instruction_table[ROL_ACCUMULATOR] = ACCUMULATOR;
	instruction_table[ROL_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[ROL_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[ROL_ABSOLUTE] = cABSOLUTE;
	instruction_table[ROL_ABSOLUTE_X] = ABSOLUTE_X;
	
	instruction_table[ROR_ACCUMULATOR] = ACCUMULATOR;
	instruction_table[ROR_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[ROR_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[ROR_ABSOLUTE] = cABSOLUTE;
	instruction_table[ROR_ABSOLUTE_X] = ABSOLUTE_X;

	instruction_table[RTI_IMPLIED] = IMPLIED;
	instruction_table[RTS_IMPLIED] = IMPLIED;

	instruction_table[SBC_IMMEDIATE] = IMMEDIATE;
	instruction_table[SBC_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[SBC_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[SBC_ABSOLUTE] = cABSOLUTE;
	instruction_table[SBC_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[SBC_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[SBC_INDIRECT_X] = INDIRECT_X;
	instruction_table[SBC_INDIRECT_Y] = INDIRECT_Y;
	
	instruction_table[SEC_IMPLIED] = IMPLIED;
	instruction_table[SED_IMPLIED] = IMPLIED;
	instruction_table[SEI_IMPLIED] = IMPLIED;	
	
	instruction_table[STA_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[STA_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[STA_ABSOLUTE] = cABSOLUTE;
	instruction_table[STA_ABSOLUTE_X] = ABSOLUTE_X;
	instruction_table[STA_ABSOLUTE_Y] = ABSOLUTE_Y;
	instruction_table[STA_INDIRECT_X] = INDIRECT_X;
	instruction_table[STA_INDIRECT_Y] = INDIRECT_Y;

	instruction_table[STX_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[STX_ZERO_PAGE_Y] = ZERO_PAGE_Y;
	instruction_table[STX_ABSOLUTE] = cABSOLUTE;
	
	instruction_table[STY_ZERO_PAGE] = ZERO_PAGE;
	instruction_table[STY_ZERO_PAGE_X] = ZERO_PAGE_X;
	instruction_table[STY_ABSOLUTE] = cABSOLUTE;
	
	instruction_table[TAX_IMPLIED] = IMPLIED;
	instruction_table[TAY_IMPLIED] = IMPLIED;
	instruction_table[TSX_IMPLIED] = IMPLIED;
	instruction_table[TXA_IMPLIED] = IMPLIED;
	instruction_table[TXS_IMPLIED] = IMPLIED;
	instruction_table[TYA_IMPLIED] = IMPLIED;

	// Set the interrupt line to zero by default
	mintInterruptLine = 0;

	mlngCycleCounter = 0;	
}





//////////////////////////////////////////////////////////////////////
// Procedure: cpu6502 Deconstructor
//////////////////////////////////////////////////////////////////////
// Description:  In charge of freeing all system memory
//////////////////////////////////////////////////////////////////////
cpu6502::~cpu6502()
{	
	// Free memory 
	if (memory != NULL)
	{
		delete memory;
		memory = NULL;
	}
}



//////////////////////////////////////////////////////////////////////
// Procedure: Init
//////////////////////////////////////////////////////////////////////
// Description: Initializes all register values, allocates memory. 
//////////////////////////////////////////////////////////////////////
void cpu6502::Init()
{
	// Provide initial values for registers
	reg.acc = 0;
	reg.x = 0;
	reg.y = 0;
	reg.pc = 0;
	reg.sp = 0;	
	reg.status = SET_UNUSED;

	// Free memory if needed (in case of a system reset)
	if (memory != NULL)
	{
		delete memory;
	}
	// Allocate some memory.  NOTE: this amount should be 
	// changed in derived classes
	memory = new unsigned char[0xFF];
}



//////////////////////////////////////////////////////////////////////
// Procedure: handleInterruptRequest
//////////////////////////////////////////////////////////////////////
// Description:  Provides a model for an interrupt handler.  Different
// machines have a different interrupts, but the general idea of pushing
// the status register and the PC to the stack and fetching a new PC
// from a memory vector should be the same.
//////////////////////////////////////////////////////////////////////
void cpu6502::handleInterruptRequest()
{	
	// Standard interrupt opening: push PC and status onto stack
	memory[reg.sp + 0x100] = ((reg.pc + 2) & 0xFF00) >> 8;	// PCH to reg.sp
	memory[reg.sp - 1 + 0x100] = (reg.pc + 2) & 0xFF;       // PCL to reg.sp-1				
	memory[reg.sp - 2 + 0x100] = reg.status;				// Status to reg.sp-2	

	reg.sp = reg.sp - 3;									// Adjust stack pointer
	
	// Now load the new PC vector from memory depending on which interrupt was received
	switch (mintInterruptLine)
	{
	//case BRK:
	case IRQ:		
		reg.pc = memory[0xFFFE] + (memory[0xFFFF] << 8);

	case NMI:		
		reg.pc = memory[0xFFFA] + (memory[0xFFFB] << 8);
	}

	// At this point, execution should resume until a RTI instruction is encountered
	mintInterruptLine = 0;

	// Process of entering an interrupt takes 7 cycles
	mlngCycleCounter += 7;
}



//////////////////////////////////////////////////////////////////////
// Procedure: RAM_Read
//////////////////////////////////////////////////////////////////////
// Description: Reads a byte from CPU memory.  This is in a separate 
// procedure because many/all computer systems use some kind of memory
// mapping scheme where you can't just look up the value directly.
//////////////////////////////////////////////////////////////////////
unsigned char cpu6502::RAM_Read(Uint16 addr)
{	
	// For our base 6502 class, we simply return the memory value at the given location
	return memory[addr];	
}



//////////////////////////////////////////////////////////////////////
// Procedure: RAM_Write
//////////////////////////////////////////////////////////////////////
// Description: Write a byte to CPU memory.  See RAM_Read for more
// details.
//////////////////////////////////////////////////////////////////////
void cpu6502::RAM_Write(Uint16 addr, unsigned char data)
{
	// For the simple 6502 class just write the value
	memory[addr] = data;
}



//////////////////////////////////////////////////////////////////////
// Procedure: executeInstruction
//////////////////////////////////////////////////////////////////////
// Description: Executes a single instruction based on the contents 
// of memory and the registers.  Alters the machine state and maintains
// the cycle counter.  This procedure also checks the interrupt line
// and calls the interrupt handler if needed.
//
// This is the main procedure as far as emulating the functionality of
// the 6502 goes.  It should probably also either be inlined or otherwise
// re-worked as it is likely to be called a LOT of times, and the repeated
// function calling might/will get expensive.
//////////////////////////////////////////////////////////////////////
void cpu6502::executeInstruction()
{			
	// ----------
	// Check for interrupts
	if (mintInterruptLine) {
		if (reg.status & SET_INTERRUPT) {			
			// If the interrupt flag of reg.status is set, then only let NMI interrupts
			// pass through
			if (mintInterruptLine & NMI) {
				mintInterruptLine = NMI;
				handleInterruptRequest();
				return;
			}
		} else {
			handleInterruptRequest();
			return;
		}				
	}
	// -----------

	unsigned int temp, temp2;	// Temporary integers

	int op = memory[reg.pc];
	//int op = RAM_Read(reg.pc);


	// Accumulate clock cycles for timing synchronization.  We add 2 here because at the least,
	// most instructions use 2 instructions.  Later, we add more cycles for various addressing
	// modes and special cases (i.e. page crossings)
	mlngCycleCounter += 2;

	// Calculate the memory index for instruction based on addressing mode into temp2	
	switch( instruction_table[op] ) {
	case IMMEDIATE:
		temp2 = reg.pc + 1;
		reg.pc += 2;
		break;
	case ZERO_PAGE:
		temp2 = memory[reg.pc+1];
		reg.pc += 2;
		mlngCycleCounter += 1;
		break;
	case ZERO_PAGE_X:				
		temp2 = (memory[reg.pc+1] + reg.x) & 0xFF;
		reg.pc += 2;
		mlngCycleCounter += 2;
		break;
	case ZERO_PAGE_Y:		
		temp2 = (memory[reg.pc+1] + reg.y) & 0xFF;
		reg.pc += 2;
		mlngCycleCounter += 2;
		break;
	case INDIRECT:
	case cABSOLUTE:		
		temp2 = memory[reg.pc+1] + (memory[reg.pc+2] << 8);
		reg.pc += 3;
		mlngCycleCounter += 2;	
		break;
	case ABSOLUTE_X:
		temp2 = memory[reg.pc+1] + (memory[reg.pc+2] << 8) + reg.x;
		reg.pc += 3;
		mlngCycleCounter += 2;
		
		// Add 1 cycle on page crossing except for these instructions:
		switch (op) {
		case ASL_ABSOLUTE_X:
		case DEC_ABSOLUTE_X:
		case INC_ABSOLUTE_X:
		case LSR_ABSOLUTE_X:
		case ROL_ABSOLUTE_X:
		case ROR_ABSOLUTE_X:
			break;
		default:
			if ( ((temp2 - reg.x) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
			break;
		}
		break;
	case ABSOLUTE_Y:
		temp2 = memory[reg.pc+1] + (memory[reg.pc+2] << 8) + reg.y;
		reg.pc += 3;
		mlngCycleCounter += 2;		
		if ( ((temp2 - reg.y) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;			
		break;

	case INDIRECT_X:		
		temp2 = memory[ (memory[reg.pc+1] + reg.x) & 0xFF ] + (memory[ ((memory[reg.pc+1] + reg.x) & 0xFF) + 1 ] << 8);
		reg.pc += 2;
		mlngCycleCounter += 4;
		break;

	case INDIRECT_Y:
		// Changed 8.29.02				
		//      .... zero page byte 1 ....   ..... zero page byte 2  ............   y reg
		temp2 = memory[ memory[reg.pc+1] ] + (memory[ memory[reg.pc+1] + 1] << 8) + reg.y;
		reg.pc += 2;
		mlngCycleCounter += 3;
		// Add 1 cycle on page crossing except for this instruction
		if (op != STA_INDIRECT_Y) {
			if ( ((temp2 - reg.y) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;			
		}
		break;	
	default:
		break;
	}
		
			
	// Execute Instruction
	switch (op) {
	// Add 1 for page boundary crossings for special modes
	case ADC_INDIRECT_Y:
	case ADC_ABSOLUTE_X:
	case ADC_ABSOLUTE_Y:
		//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
	case ADC_IMMEDIATE:
	case ADC_ZERO_PAGE:
	case ADC_ZERO_PAGE_X:
	case ADC_ABSOLUTE:	
	case ADC_INDIRECT_X:	
			// Added 9.5.02
			temp = reg.acc + memory[temp2];
			if (reg.status & SET_CARRY) temp++;			
			//reg.status &= CLEAR_ZERO;
			//reg.status &= CLEAR_OVERFLOW;
			//reg.status &= CLEAR_CARRY;
			//reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZOCS;
			if (!(temp & 0xFF))	reg.status |= SET_ZERO;									
			if (temp > 0xFF) reg.status |= SET_CARRY;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			/*if (  (temp > 0xFF) ^ (temp & 0x80) )	// Overflow = XOR of carry and sign
			{										// from nesdev
				reg.status |= SET_OVERFLOW;
			}*/
			if (!((reg.acc ^ memory[temp2]) & 0x80) && ((reg.acc ^ temp) & 0x80)) {
				reg.status |= SET_OVERFLOW;
			}
			reg.acc = (temp & 0xFF);			
			break;
	
	case AND_ABSOLUTE_X:
	case AND_ABSOLUTE_Y:
	case AND_INDIRECT_Y:
		//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
	case AND_IMMEDIATE:
	case AND_ZERO_PAGE:
	case AND_ZERO_PAGE_X:
	case AND_ABSOLUTE:	
	case AND_INDIRECT_X:	
			reg.acc &= memory[temp2];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc == 0)	reg.status |= SET_ZERO;
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;			
			break;
	
	case ASL_ACCUMULATOR:
			reg.status &= CLEAR_CARRY;
			if (reg.acc & 0x80) reg.status |= SET_CARRY;
			
			// Modified 9.4.02
			reg.acc = (reg.acc << 1) & 0xFF;
			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.acc == 0) reg.status |= SET_ZERO;		
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;					
			reg.pc++;			

			break;

	case ASL_ABSOLUTE_X:
			mlngCycleCounter++;
	case ASL_ZERO_PAGE:
	case ASL_ZERO_PAGE_X:
	case ASL_ABSOLUTE:
			mlngCycleCounter += 2;
			temp = memory[temp2];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			reg.status &= CLEAR_CARRY;
			if (temp & 0x80) reg.status |= SET_CARRY;
			
			temp = (temp << 1) & 0xFF;
						
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (temp == 0) reg.status |= SET_ZERO;		
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			//memory[ temp2 ] = temp;
			RAM_Write(temp2, temp);
			break;
		
	case BCC_RELATIVE:			
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_CARRY))
			{				
				if (memory[reg.pc-1] & 0x80)
				{					
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			}
			break;			
	
	case BCS_RELATIVE:						
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement
			if ((reg.status & SET_CARRY))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;			
		
	case BEQ_RELATIVE:						
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement						
			if ((reg.status & SET_ZERO)) {				
				if (memory[reg.pc-1] & 0x80) {
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 			
			break;
		
	case BIT_ABSOLUTE:
	case BIT_ZERO_PAGE:			
			temp = RAM_Read(temp2);
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_OVERFLOW;
			reg.status &= CLEAR_ZERO;*/
			reg.status &= CLEAR_SOZ;
			if ( (reg.acc & temp) == 0) reg.status |= SET_ZERO;
			if (temp & 0x80) reg.status |= SET_SIGN;
			if (temp & 0x40) reg.status |= SET_OVERFLOW;		
			break;

	case BMI_RELATIVE:			
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement			
			if ((reg.status & SET_SIGN))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BNE_RELATIVE:									
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_ZERO)) {
				if (memory[reg.pc-1] & 0x80) {
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else	{
					reg.pc += memory[reg.pc-1];
				}				
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			}
			break;

		case BPL_RELATIVE:
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_SIGN))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BRK_IMPLIED:
			// Trigger a BRK on the interrupt line						
			/*reg.status |= SET_BRK;
			mintInterruptLine |= BRK;
			reg.pc += 2;	// TODO: Some people may argue this ought to be 2*/

			reg.pc += 2;
			memory[reg.sp + 0x100] = ((reg.pc) & 0xFF00) >> 8;	// PCH to reg.sp
			memory[reg.sp - 1 + 0x100] = (reg.pc) & 0xFF;       // PCL to reg.sp-1							
			memory[reg.sp - 2 + 0x100] = reg.status | SET_BRK;		// Status to reg.sp-2, plus BRK flag	
			reg.sp = reg.sp - 3;									// Adjust stack pointer

			reg.status |= SET_INTERRUPT;			
			reg.pc = memory[0xFFFE] + (memory[0xFFFF] << 8);
			mlngCycleCounter += 5;								// BRK takes 7 cycles, need 5 more
			break;

		case BVC_RELATIVE:			
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement
			if (!(reg.status & SET_OVERFLOW))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BVS_RELATIVE:
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if ((reg.status & SET_OVERFLOW))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case CLC_IMPLIED:			
			reg.status &= CLEAR_CARRY;
			reg.pc++;
			break;
		
		case CLD_IMPLIED:			
			reg.status &= CLEAR_DECIMAL;
			reg.pc++;
			break;
		
		case CLI_IMPLIED:			
			reg.status &= CLEAR_INTERRUPT;
			reg.pc++;
			break;
		
		case CLV_IMPLIED:			
			reg.status &= CLEAR_OVERFLOW;
			reg.pc++;
			break;
		
		case CMP_INDIRECT_Y:
		case CMP_ABSOLUTE_X:
		case CMP_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;		
		case CMP_IMMEDIATE:
		case CMP_ZERO_PAGE:
		case CMP_ZERO_PAGE_X:
		case CMP_ABSOLUTE:		
		case CMP_INDIRECT_X:		
			temp = reg.acc - memory[temp2]; 
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;			
			if (temp < 0x100) reg.status |= SET_CARRY;
			
			break;
		
		case CPX_IMMEDIATE:
		case CPX_ZERO_PAGE:
		case CPX_ABSOLUTE:
			temp = reg.x - memory[temp2];
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;
			if (temp < 0x100) reg.status |= SET_CARRY;
			
			break;
		
		case CPY_IMMEDIATE:
		case CPY_ZERO_PAGE:
		case CPY_ABSOLUTE:
			temp = reg.y - memory[temp2];
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;
			if (temp < 0x100) reg.status += SET_CARRY;					
			break;
		
		case DEC_ABSOLUTE_X:
			mlngCycleCounter++;		
		case DEC_ZERO_PAGE:
		case DEC_ZERO_PAGE_X:
		case DEC_ABSOLUTE:					
			mlngCycleCounter += 2;
			temp = (memory[ temp2 ] - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (temp == 0) reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			//memory[ temp2 ] = temp;
			RAM_Write(temp2, temp+1);	// Write-back unmodified value
			RAM_Write(temp2, temp);
			break;

		case DEX_IMPLIED:			
			reg.x = (reg.x - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.x == 0) reg.status += SET_ZERO;			
			if (reg.x & 0x80) reg.status += SET_SIGN;						
			reg.pc++;
			break;

		case DEY_IMPLIED:
			reg.y = (reg.y - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.y == 0) reg.status |= SET_ZERO;			
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			reg.pc++;
			break;
		
		case EOR_INDIRECT_Y:
		case EOR_ABSOLUTE_X:
		case EOR_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case EOR_IMMEDIATE:
		case EOR_ZERO_PAGE:
		case EOR_ZERO_PAGE_X:
		case EOR_ABSOLUTE:		
		case EOR_INDIRECT_X:
		
			reg.acc ^= memory[temp2];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (reg.acc == 0) reg.status |= SET_ZERO;					
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;					
			break;

		case INC_ABSOLUTE_X:
			mlngCycleCounter++;
		case INC_ZERO_PAGE:
		case INC_ZERO_PAGE_X:
		case INC_ABSOLUTE:
			mlngCycleCounter += 2;
			temp = (memory[ temp2 ] + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (temp == 0) reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			RAM_Write(temp2, temp-1);	// Write-back unmodified value
			RAM_Write(temp2, temp);
			break;

		case INX_IMPLIED:
			reg.x = (reg.x + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (!reg.x) reg.status |= SET_ZERO;			
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;

		case INY_IMPLIED:			
			reg.y = (reg.y + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (!reg.y) reg.status |= SET_ZERO;			
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;

		case JMP_ABSOLUTE:			
			mlngCycleCounter++;
			reg.pc = temp2;
			break;

		case JMP_INDIRECT:	
			mlngCycleCounter += 3;
			// CPU bug:  address gets wrapped over page
			reg.pc = memory[temp2] + (memory[temp2 + 1] << 8);
			//reg.pc = memory[temp2] + (memory[ (temp2 & 0xFF00) + ((temp2 + 1) & 0xFF)] << 8);
			break;

		case JSR_ABSOLUTE:			
			memory[reg.sp + 0x100] = ((reg.pc + 2) & 0xFF00) >> 8;	// PCH to reg.sp
			memory[reg.sp - 1 + 0x100] = (reg.pc + 2) & 0xFF;       // PCL to reg.sp-1
			reg.sp -= 2;											// Sub 2 since pc is 16 bits			
			temp = memory[reg.pc + 1] + (memory[reg.pc + 2] << 8);  // New PC is absolute operand
			reg.pc = temp;
			mlngCycleCounter += 2;									// JSR = 6 clock cycles, need 2 more
			break;
	
		case LDA_ABSOLUTE_X:
		case LDA_ABSOLUTE_Y:
		case LDA_INDIRECT_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDA_IMMEDIATE:
		case LDA_ZERO_PAGE:
		case LDA_ZERO_PAGE_X:
		case LDA_ABSOLUTE:		
		case LDA_INDIRECT_X:		
			reg.acc = RAM_Read(temp2) & 0xFF;			
			/*reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;*/
			reg.status &= CLEAR_ZS;
			if (reg.acc == 0) reg.status |= SET_ZERO;					
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;								
			break;
		
		
		case LDX_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDX_IMMEDIATE:
		case LDX_ZERO_PAGE:
		case LDX_ZERO_PAGE_Y:
		case LDX_ABSOLUTE:					
			reg.x = RAM_Read(temp2) & 0xFF;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.x == 0)	reg.status |= SET_ZERO;						
			if (reg.x & 0x80) reg.status |= SET_SIGN;							
			break;
		
		case LDY_ABSOLUTE_X:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
			if ( (temp2 - reg.x) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDY_IMMEDIATE:
		case LDY_ZERO_PAGE:
		case LDY_ZERO_PAGE_X:
		case LDY_ABSOLUTE:					
			reg.y = RAM_Read(temp2) & 0xFF;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (reg.y == 0)	reg.status |= SET_ZERO;					
			if (reg.y & 0x80) reg.status |= SET_SIGN;					
			break;
	
		case LSR_ACCUMULATOR:
			// Added 9.4.02			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x1) reg.status |= SET_CARRY;
			
			reg.acc = reg.acc >> 1;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;
	
		case LSR_ABSOLUTE_X:
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case LSR_ZERO_PAGE:			
		case LSR_ZERO_PAGE_X:
		case LSR_ABSOLUTE:		
			// Added 9.4.02			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;			
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			if (temp & 0x1) reg.status |= SET_CARRY;
			
			temp = temp >> 1;
			if (temp == 0) reg.status |= SET_ZERO;									
			//memory[ temp2 ] = temp;						
			RAM_Write(temp2, temp);
			break;
		
		case NOP_IMPLIED:
			reg.pc++;
			break;
		
		case ORA_ABSOLUTE_X:
		case ORA_ABSOLUTE_Y:
		case ORA_INDIRECT_Y:			
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case ORA_IMMEDIATE:
		case ORA_ZERO_PAGE:
		case ORA_ZERO_PAGE_X:
		case ORA_ABSOLUTE:		
		case ORA_INDIRECT_X:
			reg.acc |= memory[ temp2 ];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;						
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;						
			break;
		
		case PHA_IMPLIED:			
			memory[reg.sp + 0x100] = reg.acc;
			reg.sp--;
			reg.pc++;
			mlngCycleCounter += 1;					// PHA = 3 cycles, need 1 more
			break;
		
		case PHP_IMPLIED:			
			// BRK flag apparently gets pushed as being set no matter what here
			memory[reg.sp + 0x100] = (reg.status | SET_BRK);
			reg.sp--;
			reg.pc++;
			mlngCycleCounter += 1;
			break;

		case PLA_IMPLIED:			
			reg.sp++;
			reg.acc = memory[reg.sp + 0x100];			
			reg.pc++;
			
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			
			mlngCycleCounter += 2;
			break;
		
		case PLP_IMPLIED:
			reg.sp++;
			reg.status = memory[reg.sp + 0x100] | SET_UNUSED;		
			reg.pc++;
			mlngCycleCounter += 2;
			break;
		
		case ROL_ACCUMULATOR:
			reg.acc = (reg.acc << 1);
			if (reg.status & SET_CARRY) reg.acc |= 0x01;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc > 0xFF) reg.status |= SET_CARRY;
		
			reg.acc &= 0xFF;
			
			if (reg.acc == 0) reg.status |= SET_ZERO;			
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
	
			reg.pc++;
			break;

		case ROL_ABSOLUTE_X:			
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case ROL_ZERO_PAGE:
		case ROL_ZERO_PAGE_X:
		case ROL_ABSOLUTE:		
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			temp = (temp << 1);
			if (reg.status & SET_CARRY) temp |= 0x01;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (temp > 0xFF) reg.status |= SET_CARRY;						
			temp &= 0xFF;
			if (temp == 0)  reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			//memory[ temp2 ] = temp;			
			RAM_Write(temp2, temp);
			break;
		
		case ROR_ACCUMULATOR:						
			if (reg.status & SET_CARRY) reg.acc |= 0x100;
			if (reg.acc & 1)
			{
				reg.status |= SET_CARRY;
			} else {
				reg.status &= CLEAR_CARRY;
			}
			
			reg.acc = (reg.acc >> 1);
			
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
																
			if (reg.acc == 0) reg.status |= SET_ZERO;			
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;
			
		case ROR_ABSOLUTE_X:
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case ROR_ZERO_PAGE:
		case ROR_ZERO_PAGE_X:
		case ROR_ABSOLUTE:		
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			if (reg.status & SET_CARRY) temp |= 0x100;
			if (temp & 1)
			{
				reg.status |= SET_CARRY;
			} else	{
				reg.status &= CLEAR_CARRY;
			}
			
			temp = (temp >> 1);

			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;			

			if (temp == 0) reg.status += SET_ZERO;			
			if (temp & 0x80) reg.status += SET_SIGN;
			
			//memory[ temp2 ] = temp;			
			RAM_Write(temp2, temp);
			break;

		case RTI_IMPLIED:			
			reg.sp++;
			reg.status = memory[reg.sp + 0x100];				// Pull off status
			reg.sp++;
			reg.pc = memory[reg.sp + 0x100];					// Pull off PCL
			reg.sp++;
			reg.pc = reg.pc + (memory[reg.sp + 0x100] << 8);	// Pull off PCH
			mlngCycleCounter += 4;								// RTI = 6 cycles, need 4 more			
			break;

		case RTS_IMPLIED:
			reg.sp += 2;
			reg.pc = memory[reg.sp - 1 + 0x100];
			reg.pc = reg.pc + (memory[reg.sp + 0x100] << 8);						
			reg.pc++;
			mlngCycleCounter += 4;						// RTS = 6 cycles, need 4 more
			break;
       
        case SBC_ABSOLUTE_X:
		case SBC_ABSOLUTE_Y:
		case SBC_INDIRECT_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case SBC_IMMEDIATE:
		case SBC_ZERO_PAGE:
		case SBC_ZERO_PAGE_X:
		case SBC_ABSOLUTE:
		case SBC_INDIRECT_X:		
			temp = reg.acc - memory[ temp2 ];
			if (!(reg.status & SET_CARRY)) temp--;
			
			/*reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_OVERFLOW;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_SIGN;*/
			
			reg.status &= CLEAR_ZOCS;
			if (!(temp & 0xFF))	reg.status |= SET_ZERO;
								
			/*if (  (temp > 0xFF) ^ (temp & 0x80) )	// Overflow = XOR of carry and sign
			{										// from nesdev
				reg.status |= SET_OVERFLOW;
			}*/
			if (((reg.acc ^ memory[temp2]) & 0x80) && ((reg.acc ^ temp) & 0x80)) {
				reg.status |= SET_OVERFLOW;
			}


			if (temp < 0x100) reg.status |= SET_CARRY;		
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			reg.acc = (temp & 0xFF);			
			break;
	
		case SEC_IMPLIED:
			reg.status &= CLEAR_CARRY;
			reg.status += SET_CARRY;
			reg.pc++;
			break;

		case SED_IMPLIED:
			reg.status &= CLEAR_DECIMAL;
			reg.status += SET_DECIMAL;
			reg.pc++;
			break;

		case SEI_IMPLIED:
			reg.status &= CLEAR_INTERRUPT;
			reg.status += SET_INTERRUPT;
			reg.pc++;
			break;
		
		case STA_ABSOLUTE_X:
		case STA_ABSOLUTE_Y:
		case STA_INDIRECT_Y:
			mlngCycleCounter++;
		case STA_ZERO_PAGE:
		case STA_ZERO_PAGE_X:
		case STA_ABSOLUTE:		
		case STA_INDIRECT_X:
		
			RAM_Write(temp2, reg.acc);
			//memory[ temp2 ] = reg.acc;			
			break;
		
		case STX_ZERO_PAGE:
		case STX_ZERO_PAGE_Y:
		case STX_ABSOLUTE:
			RAM_Write(temp2, reg.x);
			//memory[ temp2 ] = reg.x;			
			break;

		case STY_ZERO_PAGE:
		case STY_ZERO_PAGE_X:
		case STY_ABSOLUTE:
			RAM_Write(temp2, reg.y);
			//memory[ temp2 ] = reg.y;			
			break;
		
		case TAX_IMPLIED:			
			reg.x = reg.acc;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			if (reg.x == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TAY_IMPLIED:			
			reg.y = reg.acc;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			if (reg.y == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TSX_IMPLIED:			
			// fixed 9.13.02
			reg.x = reg.sp;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			if (reg.x == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TXA_IMPLIED:			
			reg.acc = reg.x;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;
		
		case TXS_IMPLIED:			
			reg.sp = reg.x;			
			reg.pc++;
			break;

		case TYA_IMPLIED:		
			reg.acc = reg.y;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;		
		
		default:
			reg.pc++;
	}		
}

























