//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//////////////////////////////////////////////////////////////////////
// cpu6502.h: interface for the cpu6502 class.
//////////////////////////////////////////////////////////////////////
// Class Description:  The cpu6502 class emulates the registers and opcodes of a 
// 6502 processor.  Many aspects of the system must be filled in by derived classes.
// For example: the size of memory, the loading of the program code, interrupt handlers,
// and other things of that nature.
//////////////////////////////////////////////////////////////////////

#ifndef CPU6502
#define CPU6502

#include "Porting/basetypes.h"
#include "Opcodes.h"	// Include constants for all op_code


// Bitmasks used to set various status fields
#define		SET_CARRY			0x01
#define		SET_ZERO			0x02
#define		SET_INTERRUPT		0x04
#define		SET_DECIMAL			0x08
#define		SET_BRK				0x10
#define		SET_UNUSED			0x20
#define		SET_OVERFLOW		0x40
#define		SET_SIGN			0x80

#define		CLEAR_CARRY			0xFE
#define		CLEAR_ZERO			0xFD
#define		CLEAR_INTERRUPT		0xFB
#define		CLEAR_DECIMAL		0xF7
#define		CLEAR_BRK			0xEF
#define		CLEAR_UNUSED		0xDF
#define		CLEAR_OVERFLOW		0xBF
#define		CLEAR_SIGN			0x7F


#define		CLEAR_SZC			CLEAR_SIGN & CLEAR_ZERO & CLEAR_CARRY
#define		CLEAR_SOZ			CLEAR_SIGN & CLEAR_OVERFLOW & CLEAR_ZERO
#define		CLEAR_ZS			CLEAR_ZERO & CLEAR_SIGN
#define		CLEAR_ZOCS			CLEAR_ZERO & CLEAR_OVERFLOW & CLEAR_CARRY & CLEAR_SIGN

// Constants used for Interrupt line
#define		BRK					0x01
#define		IRQ					0x02
#define		NMI					0x04
#define		RESET				0x08
#define		APUIRQ				0x10


class cpu6502  
{
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	cpu6502();
	virtual ~cpu6502();	
	virtual void Init();
	virtual void executeInstruction();
	virtual unsigned char RAM_Read(Uint16 addr);
	virtual void RAM_Write(Uint16 addr, unsigned char data);
	virtual void handleInterruptRequest();


	///////////////////////////////
	// Class Data Members
	///////////////////////////////

	// Instruction Table
	// Format:  instruction_table[opcode] = addressing mode
	unsigned char instruction_table[256];
		
	// Register struct
	struct regs {
	    // These registers are declared as ints (more than 8 bits) because we have to be able to
		// tell if an overflow occurred, and the way I do that is by checking bit 9
		unsigned int	 acc;		// Accumlator (8 bits)
		unsigned int	 x;			// X (8 bits)
		unsigned int	 y;	  		// Y (8 bits)
		unsigned int	 status;	// Status register (8 bits)
		unsigned int	 pc;		// Program Counter (16 bits)
		unsigned char	 sp;		// Stack Pointer (8 bits) (declared as unsigned 8bits because
		 		                    // overflow math is needed for nearly ever operation on reg.sp
	};

	regs reg;
	
	unsigned char *memory;				// Memory Pointer	
	unsigned long mlngCycleCounter;		// Cycle counter (32 bit)	
	unsigned int mintInterruptLine;		// Interrupt line		

	private:		
};

#endif
