//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes Mapper 4.cpp - Implementation of mapper #4 (MMC3)
//////////////////////////////////////////////////////////////////////

#include "NesMapper004.h"


NesMapper4::NesMapper4(NesCartridge *linkCartridge, NesCPU *&linkCPU, NesPPU *linkPPU) {
	mCartridge = linkCartridge;	
	mPpu = linkPPU;
			
	mCpu = linkCPU;
	
	// Replace the standard PPU memory object with a MMC3 specific one	
	NesPPUMemory *newPPUMemory = new MMC3PPUMemory();	
	newPPUMemory->link(this);
	newPPUMemory->init();
	newPPUMemory->mbytMirroringType = linkPPU->VRAM->mbytMirroringType;
	delete mPpu->VRAM;
	mPpu->VRAM = newPPUMemory;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	status.blnIRQsEnabled = false;
	status.bytPRGSwitchSlots = 0;
	status.bytIRQLatch = 0;
	
	initMaps();
	
	int int8KPRGBanks = mCartridge->header.PRGROM_count << 1;
	
	// These variables map what banks are occupied in the prg space
	bytPrg0 = 0;
	bytPrg1 = 1;
	bytPrgNextLast  = int8KPRGBanks - 2;
	bytPrgLast		= int8KPRGBanks - 1;	
	
	chr01 = 0;
	chr23 = 2;
	chr4  = 4;
	chr5  = 5;
	chr6  = 6;
	chr7  = 7;

	IrqReset = 0;
	latched = 0;
}



// On RESET/power on, the last 2 banks of 8k PRGROM are swapped into C000 and E000, while
// the first 2 8k banks are swapped into 8000 and A000
// NOTE: "c++ counting from 0" deal makes this a little more confusing than it should be
void NesMapper4::initMemory()
{

	swapPRGBank8(0x8000, bytPrg0);
	swapPRGBank8(0xA000, bytPrg1);
	swapPRGBank8(0xC000, bytPrgNextLast);
	swapPRGBank8(0xE000, bytPrgLast);
	
	if (bytCHRCount) swapCHRBank8(0);
}



void NesMapper4::write(Uint16 addr, unsigned char data) {		
	addr &= 0xE001;
	switch (addr)
	{
	// Register $8000: Select command, PRG switching areas, and CHR address
	case 0x8000:
		status.bytCommand = data & 0x07;
		status.bytPRGSwitchSlots = data & 0x40;
		status.bytCHRAddrSel = data & 0x80;		
		if (mCartridge->header.CHRROM_count) {
			if (!status.bytCHRAddrSel) {
				swapCHRBank1( 0x0000, chr01);
				swapCHRBank1( 0x0400, chr01+1);
				swapCHRBank1( 0x0800, chr23);
				swapCHRBank1( 0x0C00, chr23+1);
				swapCHRBank1( 0x1000, chr4);
				swapCHRBank1( 0x1400, chr5);
				swapCHRBank1( 0x1800, chr6);
				swapCHRBank1( 0x1C00, chr7);
			} else {
				swapCHRBank1( 0x1000, chr01);
				swapCHRBank1( 0x1400, chr01+1);
				swapCHRBank1( 0x1800, chr23);
				swapCHRBank1( 0x1C00, chr23+1);
				swapCHRBank1( 0x0000, chr4);
				swapCHRBank1( 0x0400, chr5);
				swapCHRBank1( 0x0800, chr6);
				swapCHRBank1( 0x0C00, chr7);
			}
		}
		
		// Swap PRGROM banks on a switchable slots change
		if (status.bytPRGSwitchSlots) {
			swapPRGBank8(0x8000, bytPrgNextLast);
			swapPRGBank8(0xA000, bytPrg1);
			swapPRGBank8(0xC000, bytPrg0);
			//swapPRGBank8(0xE000, bytPrgLast);
		} else {
			swapPRGBank8(0x8000, bytPrg0);
			swapPRGBank8(0xA000, bytPrg1);
			swapPRGBank8(0xC000, bytPrgNextLast);
			//swapPRGBank8(0xE000, bytPrgLast);
		}

		break;
	
		// Register $8001: Activate a command from $8000
	case 0x8001:
				
		// Calculate and perform PPU switching (commands 0-5)
		if ((status.bytCommand <= 0x05) && (bytCHRCount)) {	
			// We need to keep track of what CHR-ROM banks are where in 
			// case the swapping space gets fillped by 8000.7
			switch (status.bytCommand) {
			case 0:
				chr01 = (data & 0xFE) % (bytCHRCount << 3);
				break;
			case 1:
				chr23 = (data & 0xFE) % (bytCHRCount << 3);
				break;
			case 2:
				chr4 = data % (bytCHRCount << 3);
				break;
			case 3:
				chr5 = data % (bytCHRCount << 3);
				break;
			case 4:
				chr6 = data % (bytCHRCount << 3);
				break;
			case 5:
				chr7 = data % (bytCHRCount << 3);
				break;
			}
			if (mCartridge->header.CHRROM_count) {
				if (!status.bytCHRAddrSel) {
					swapCHRBank1( 0x0000, chr01);
					swapCHRBank1( 0x0400, chr01+1);
					swapCHRBank1( 0x0800, chr23);
					swapCHRBank1( 0x0C00, chr23+1);
					swapCHRBank1( 0x1000, chr4);
					swapCHRBank1( 0x1400, chr5);
					swapCHRBank1( 0x1800, chr6);
					swapCHRBank1( 0x1C00, chr7);
				} else {
					swapCHRBank1( 0x1000, chr01);
					swapCHRBank1( 0x1400, chr01+1);
					swapCHRBank1( 0x1800, chr23);
					swapCHRBank1( 0x1C00, chr23+1);
					swapCHRBank1( 0x0000, chr4);
					swapCHRBank1( 0x0400, chr5);
					swapCHRBank1( 0x0800, chr6);
					swapCHRBank1( 0x0C00, chr7);
				}
			}
		}
		
		if (status.bytCommand == PRGFIRST) {
			bytPrg0 = data % (bytPRGCount << 1);	
		} else if (status.bytCommand == PRGLAST) {
			bytPrg1 = data % (bytPRGCount << 1);
		}
		if (status.bytCommand >= 0x06) {
			if (status.bytPRGSwitchSlots) {
				swapPRGBank8(0x8000, bytPrgNextLast);
				swapPRGBank8(0xA000, bytPrg1);
				swapPRGBank8(0xC000, bytPrg0);				
			} else {
				swapPRGBank8(0x8000, bytPrg0);
				swapPRGBank8(0xA000, bytPrg1);
				swapPRGBank8(0xC000, bytPrgNextLast);
			}
		}
		break;


	case 0xA000:		
		if (data & 0x01) {
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
		} else {
			mPpu->VRAM->mbytMirroringType = VERTICAL;			
		}
		
		break;
	case 0xA001:
		break;
	case 0xC000:
		status.bytIRQLatch = data;		
		latched = IrqReset ^ 0x01;
		if (IrqReset) status.intIRQCtr = data;
		//status.intIRQCtr = data;
		break;
	case 0xC001:
		//status.bytIRQLatch = data;
		status.intPPUCtr = 42;
		status.intIRQCtr = status.bytIRQLatch;
		break;
	case 0xE000:
		//status.intIRQCtr = status.bytIRQLatch;
		status.blnIRQsEnabled = false;
		IrqReset = 0x01;
		break;
	case 0xE001:
		status.blnIRQsEnabled = true;		
		status.intPPUCtr = 42;
		if (latched) status.intIRQCtr = status.bytIRQLatch;
		break;
	}
}





//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the mapper to the file stream
//-------------------------------------------------------------------
void NesMapper4::saveState(FILE *stream) {
	NesBaseMapper::saveState(stream);
	fwrite(&status, sizeof(status), 1, stream);		// misc
	fwrite(&bytPrg0, sizeof(bytPrg0), 1, stream);  
	fwrite(&bytPrg1, sizeof(bytPrg0), 1, stream);  
	fwrite(&bytPrgLast, sizeof(bytPrg0), 1, stream);  
	fwrite(&bytPrgNextLast, sizeof(bytPrg0), 1, stream); 
	fwrite(&chr01, sizeof(chr01), 1, stream);
	fwrite(&chr23, sizeof(chr23), 1, stream);
	fwrite(&chr4, sizeof(chr4), 1, stream);
	fwrite(&chr5, sizeof(chr5), 1, stream);
	fwrite(&chr6, sizeof(chr6), 1, stream);
	fwrite(&chr7, sizeof(chr7), 1, stream);	
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the mapper from the file stream
//-------------------------------------------------------------------
void NesMapper4::loadState(FILE *stream) {
	NesBaseMapper::loadState(stream);
	fread(&status, sizeof(status), 1, stream);		// misc
	fread(&bytPrg0, sizeof(bytPrg0), 1, stream);  
	fread(&bytPrg1, sizeof(bytPrg0), 1, stream);  
	fread(&bytPrgLast, sizeof(bytPrg0), 1, stream);  
	fread(&bytPrgNextLast, sizeof(bytPrg0), 1, stream);  
	fread(&chr01, sizeof(chr01), 1, stream);
	fread(&chr23, sizeof(chr23), 1, stream);
	fread(&chr4, sizeof(chr4), 1, stream);
	fread(&chr5, sizeof(chr5), 1, stream);
	fread(&chr6, sizeof(chr6), 1, stream);
	fread(&chr7, sizeof(chr7), 1, stream);
}