//////////////////////////////////////////////////////////////////////
// Nes Mapper 1.cpp - Implementation of mapper #1 (MMC1)
//////////////////////////////////////////////////////////////////////

#include "NesMapper1.h"




//////////////////////////////////////////////////////////////////////
// Procedure: NesMapper1Constructor
//////////////////////////////////////////////////////////////////////
// Description: Initializes the various data members of mapper #1
//////////////////////////////////////////////////////////////////////
NesMapper1::NesMapper1(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU)
{
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	// Initialize serial port
	SerialPort.bytPtr = 0;	
	SerialPort.bytBitArray = 0;
			
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	status.intBankSwitchOffset = 0x8000;
	status.blnPRGBankSize = BANKSIZE16;

	initMaps();

	regs[0] = 0x0C;
	regs[1] = regs[2] = regs[3] = 0x00;
	intLastPort = 0;
}





//////////////////////////////////////////////////////////////////////
// Procedure: initMemory
//////////////////////////////////////////////////////////////////////
// Description: On RESET/power on, the first 16k bank is loaded into
// $8000 and the last 16k bank is loaded into $C000
//////////////////////////////////////////////////////////////////////
void NesMapper1::initMemory() {		
	swapPRGBank16(0x8000, 0);
	swapPRGBank16(0xC000, bytPRGCount - 1);
	
	intBankMap16[0] = 0;					// Set up the bank map, too
	intBankMap16[1] = bytPRGCount - 1;

	if (bytCHRCount) swapCHRBank8(0);
}





//////////////////////////////////////////////////////////////////////
// Procedure: write
//////////////////////////////////////////////////////////////////////
// Description: Accepts data written from the PRGROM to the mapper.  For MMC1, this
// procedure maintains the serial port interfaces (one bit at a time)
// needed to communicate with the mapper.  Once a serial port fills a
// 5 bit array, the data is then "accepted" in a different procedure
// where the real details of MMC1 are handled.
// Modified:
//////////////////////////////////////////////////////////////////////
void NesMapper1::write(unsigned __int16 addr, unsigned char data)
{		
	int intPort = (addr & 0x6000) >> 13; // convert addr to 0,1,2 or 3

	if (data & 0x80) {
		// On a reset signal, reset count and set bits D2, D3 of reg0
		SerialPort.bytPtr = 0;
		intLastPort = intPort;		
		status.blnPRGBankSize = BANKSIZE16;
		status.intBankSwitchOffset = 0x8000;		
	} else {		
		if (intPort != intLastPort) {
			// If the port changes, reset the serial register data
			SerialPort.bytBitArray = 0;			
			SerialPort.bytPtr = 0;
			intLastPort = intPort;
		}
		
		// Accumlate the newest bit of information into serial register
		SerialPort.bytBitArray |= ((data & 0x01) << (SerialPort.bytPtr));
		SerialPort.bytPtr++;
		
		// If we reach the capacity of the register, send the data to the 
		// appropriate port
		if (SerialPort.bytPtr == 5) {
			SerialPort.bytPtr = 0;  	
    
			// Interpret the data sent to MMC1
			acceptData(SerialPort.bytBitArray, (char) intPort);
			if (intPort==0) {
				// Not sure about this...TMNT breaks if added
				//acceptData(SerialPort.bytBitArray, (char) 1);
			}
			SerialPort.bytBitArray = 0; // clear serial data
		}
	}
}





//////////////////////////////////////////////////////////////////////
// Procedure: acceptData
//////////////////////////////////////////////////////////////////////
// Description: Once a 5 bit value has been filled in (bit by bit), its
// value is accepted here.  This is where the bulk of the MMC1 code is. 
//////////////////////////////////////////////////////////////////////
void NesMapper1::acceptData(unsigned char bytData, unsigned char bytPort)
{	
	switch (bytPort)
	{		
	// 0 = $8000 to $9FFF
	case 0:
		///////////////////////// Register 0 ///////////////////////////////
		// D0 and D1 address mirroring
		// Fill in mirroring (D0 = H/V, D1 = Single Screen (override))
		switch (bytData & 0x03) {
		case 0:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2000;
			break;
		case 1:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2400;	
			break;
		case 2:
			mPpu->VRAM->mbytMirroringType = VERTICAL;
			break;
		case 3:
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
			break;
		}
		
		// D3 toggles between 16k and 32k swapping and overrides D2, if D3=0
		if (bytData & 0x08)
		{
			// Check D2 for high/low switching
			int intTemp = status.intBankSwitchOffset;
			if (bytData & 0x04)
			{
				status.intBankSwitchOffset = 0x8000;
			} else {
				status.intBankSwitchOffset = 0xC000;
			}

			if (status.blnPRGBankSize == BANKSIZE32)
			{
				trans32to16();
				status.blnPRGBankSize = BANKSIZE16;
			} 	 else {
				// If we don't switch from 32k to 16k, then look for high to low changing
				if (intTemp != status.intBankSwitchOffset) transSwitchSpaces();
			}			
		} else {
			// 32K PRG-ROM switching, ignore D2
			if (status.blnPRGBankSize == BANKSIZE16)
			{
				trans16to32();
				status.blnPRGBankSize = BANKSIZE32;
			}			
		}
			
		// Finally, D4 sets 8k or 4k CHRROM switches		
		if (bytData & 0x10)
		{
			status.blnCHRBankSize = BANKSIZE4;
		} else {
			status.blnCHRBankSize = BANKSIZE8;
		}
		break;
		
	case 1:
		///////////////////////// Register 1 ///////////////////////////////
		// Switch VROM at $0000			
		if (bytCHRCount) {
			if (status.blnCHRBankSize == BANKSIZE8)	{
				swapCHRBank8( ((bytData & 0x1F) >> 1) % bytCHRCount );
			} else if (regs[0]) {				
				swapCHRBank4(0x0000, ( ((bytData  & 0x1F)) % (bytCHRCount << 1)));
			}
		}
		break;
				
	case 2:
		///////////////////////// Register 2 ///////////////////////////////		
		// Switch VROM at $1000 - ignore if not set to 4k switching?
		// CHANGE
		if (bytCHRCount && regs[0]) {
			if (status.blnCHRBankSize == BANKSIZE4) {
				swapCHRBank4(0x1000, ((bytData  & 0x1F) % (bytCHRCount << 1)));
			}
		}
		break;
				
	case 3:
		///////////////////////// Register 3 ///////////////////////////////
		// Select ROM Bank
		if (status.blnPRGBankSize == BANKSIZE32) {
			// Swap 32k PRGROM into memory
			int intPage32 = ((bytData & 0x0F) >> 1) % (bytPRGCount >> 1);
			
			// CHANGE			
			if (regs[0]) {
				swapPRGBank32( intPage32 );
				intBankMap16[0] = intPage32 << 1;
				intBankMap16[1] = (intPage32 << 1) + 1;
			}
		} else {			
			// Swap 16k PRGROM into high or low spot			
			int intPage = (bytData & 0x0F) % bytPRGCount, intMap = (status.intBankSwitchOffset & 0x4000) >> 14;
				
			if (status.intBankSwitchOffset == 0x8000) {
				// Swap Low
				intBankMap16[ 0 ] = intPage;								
				swapPRGBank16(0x8000, intPage);				
			} else {
				// Swap High
				intBankMap16[ 0 ] = 0;				
				intBankMap16[ 1 ] = intPage;
				swapPRGBank16(0x8000, 0);
				swapPRGBank16(0xC000, intPage);
			}
		}
		blnReg3Written = true;
		break;	
	}
}



///////////////////////////////////////////////////////////////////
/////////// Transition from high to low/low to high ///////////////
// Move bank in old switchable area to new switchable area
///////////////////////////////////////////////////////////////////
void NesMapper1::transSwitchSpaces()
{	
	if (status.intBankSwitchOffset == 0x8000)
	{
		// Switched from high to low:  copy high bank to low bank						
		intBankMap16[0] = intBankMap16[1];
		intBankMap16[1] = bytPRGCount - 1;
	} else {
		// Switched from low to high - copy low bank to high bank
		intBankMap16[1] = intBankMap16[0];
		intBankMap16[0] = 0;
	}
	swapPRGBank16(0x8000, intBankMap16[0] );						
	swapPRGBank16(0xC000, intBankMap16[1] ); //copy "hardwired" bank to high bank (might nto be correct)
}




///////////////////////////////////////////////////////////////////
/////////// Transition from 16k bank to 32k bank ///////////////
// Figure out which of the 2 16k banks is the switchable one, and then
// load the 32k page that contains that 16k bank (32k banks always
// load even #s in $8000 and odd 16k #s into $C000)
///////////////////////////////////////////////////////////////////
void NesMapper1::trans16to32() {	
	int intPage32;
	if (status.intBankSwitchOffset == 0x8000)
	{
		intPage32 = intBankMap16[0];
	} else {
		intPage32 = intBankMap16[1];
	}
	intBankException = intPage32;		// Save the bank that was in switchable area for possible future use	
	intPage32 >>= 1;
	intBankMap16[0] = (intPage32 << 1);
	intBankMap16[1] = (intPage32  << 1) + 1;	
	swapPRGBank32(intPage32);
	blnReg3Written = false;
}




///////////////////////////////////////////////////////////////////
////////// Transition from 32k bank to 16k bank /////////////////
// Always take the even # bank, unless an odd bank was already there,
// got switched to 32k, and switched back to 16k without any writes to
// reg 3
///////////////////////////////////////////////////////////////////
void NesMapper1::trans32to16() {	
	int intSwitchableBank;
	
	// 1) Figure out which bank should go into switchable space
	if (blnReg3Written)
	{
		// reg 3 got a write, put even # bank in the correct place
		intSwitchableBank = intBankMap16[0];
	} else {
		// reg 3 did NOT get a write, put whichever bank was in switchable space
		// before 16->32 change back into switchable space
		intSwitchableBank = intBankException;
	}
		
	// 2) Swap the bank into the switchable space along with the "hardwired" bank
	if (status.intBankSwitchOffset == 0x8000)
	{		
		// LOW switchable		
		intBankMap16[0] = intSwitchableBank;
		intBankMap16[1] = bytPRGCount - 1;
	} else {
		// HIGH switchable
		intBankMap16[1] = intSwitchableBank;
		intBankMap16[0] = 0;
	}
	swapPRGBank16(0x8000, intBankMap16[0] );						
	swapPRGBank16(0xC000, intBankMap16[1] ); 
}
