//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes.cpp: implementation of the Nes class.
//////////////////////////////////////////////////////////////////////

#include "Nes.h"

//////////////////////////////////////////////////////////////////////
// Procedure: Nes Constructor
//////////////////////////////////////////////////////////////////////
// Description: The Nes Constructor is responsible for innitializing all
// of the components of the NES.
//////////////////////////////////////////////////////////////////////
Nes::Nes()
{
	// Allocate the CPU
	mCpu = new NesCPU();
	mCpu->Init();

	// Allocate the PPU
	mPpu = new NesPPU();
	mPpu->Init();

	// Allocate the APU
	mApu = new NesAPU();	

	mApu->linkCPU(mCpu);	// APU has to be able to fire IRQs at CPU
			
	mController1 = new NesInput(CONTROLLER1);
	mController2 = new NesInput(CONTROLLER2);
	
	// Link the PPU object to the CPU object (Note, this may be bad programming practice but it provides a way for
	// the two objects to interact and still only be data members of the Nes class)
	mCpu->linkCPU(mPpu, mApu, mController1, mController2, mMapper);	

	// By default, create an NTSC machine
	mCpu->setMachineMode(NTSCNESMODE);

	// Initialize APU now that the CPU is ready to go and has good timing values
	mApu->Init();

	mCartridge = NULL;		
	mMapper = NULL;
	
	mblnSRAMUsed = false;
	mblnCartLoaded = false;
}



//////////////////////////////////////////////////////////////////////
// Procedure: Nes Deconstructor
//////////////////////////////////////////////////////////////////////
// Description: The Nes Deconstructor is responsible for deleting all
// of the NES components.
//////////////////////////////////////////////////////////////////////
Nes::~Nes() {
	// Get rid of data object pointer	
	if (mCartridge != NULL)	delete mCartridge;			
	if (mCpu != NULL) delete mCpu;		
	if (mPpu != NULL) delete mPpu;	
	if (mApu != NULL) delete mApu;
	if (mController1 != NULL) delete mController1;
	if (mController2 != NULL) delete mController2;
	mPpu = NULL;
	mCpu = NULL;
	mApu = NULL;
	mCartridge = NULL;
}





//////////////////////////////////////////////////////////////////////
// Procedure: loadCartridge
//////////////////////////////////////////////////////////////////////
// Description: This procedure accepts a NES file as a parameter.  It
// then loads the file into the cartridge object, initializes the PPU
// object, dynamically creates the proper Mapper object, and initializes
// the CPU and PPU via the Memory Mapper.
//////////////////////////////////////////////////////////////////////
int Nes::loadCartridge(char *filename) {
	// Load file contents into memory
	FILE *stream = fopen(filename, "rb");
	if (stream == NULL) return -1;
	fseek(stream, 0, SEEK_END);		// go to the end of file
	int fileSize = ftell(stream);   // return current position
	fseek(stream, 0, 0);			// go back to start of file
	unsigned char *cartdata = new unsigned char[fileSize+1];
	fread(&cartdata[0], 1, fileSize, stream);
	fclose(stream);

	// Call the overloaded loadCartridge method to do the rest...
	int retval = loadCartridge(cartdata, filename);	
		
	// Don't need this anymore...
	delete cartdata;

	return retval;
}




//-------------------------------------------------------------------
// Procedure: loadCartridge
//-------------------------------------------------------------------
// Description: This provides another way of loading a cartridge; it reads
// the data directly from a memory buffer rather than a file.  This is useful
// if you want to load a NES file from, say, a .ZIP file and don't want to
// unzip the NES file to disk, load the cart from disk, and delete the file
//
// A filename parameter is still supplied for naming SRAM files, states, etc,
// but it is not used in the actual loading process
//-------------------------------------------------------------------
int Nes::loadCartridge(unsigned char *&cartdata, char *filename) {
	// Deconstruct current cartridge object if neccessary
	if (mCartridge != NULL)	delete mCartridge;

	// Construct a catridge with a pointer to the file contents
	mCartridge = new NesCartridge(cartdata);

	
	// Send contents of NES file to the NES object if the load was valid
	if (mCartridge->mintErrorCode == OK) {	
		// Set the proper mirroring scheme in the PPU VRAM class
		if (mCartridge->header.ROM_Control_1 & 0x01)
		{
			mPpu->VRAM->mbytMirroringType = VERTICAL;
		} else	{
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
		}
		if (mCartridge->header.ROM_Control_1 & 0x08)
		{
			mPpu->VRAM->mbytMirroringType = FOURSCREEN;
		}
		
		// Enable SRAM
		if(mCartridge->header.ROM_Control_1 & 0x02)
		{
			mblnSRAMUsed = true;
		}

		// Allocate the proper mapper object to match the cartridge
		switch (mCartridge->header.mapper)
		{
		case 0:	// mapper #0
			mMapper = new NesBaseMapper(mCartridge, mCpu, mPpu);
			break;
		case 1: // mapper #1 - MMC1
			mMapper = new NesMapper1(mCartridge, mCpu, mPpu);
			break;
		case 2:	// mapper #2 - UNROM
			mMapper = new NesMapper2(mCartridge, mCpu, mPpu);
			break;
		case 3: // mapper #3 - CNROM
			mMapper = new NesMapper3(mCartridge, mCpu, mPpu);
			break;
		case 4: // mapper #4 - MMC3
			mMapper = new NesMapper4(mCartridge, mCpu, mPpu);
			break;
		case 5: // mapper #5 - MMC5
			mMapper = new NesMapper5(mCartridge, mCpu, mPpu);
			break;
		case 6: // mapper #6 - FFE F4xxx
			mMapper = new NesMapper6(mCartridge, mCpu, mPpu);
			break;
		case 7: // mapper #7 - AOROM
			mMapper = new NesMapper7(mCartridge, mCpu, mPpu);
			break;
		case 9: // mapper #9 - MMC2
			mMapper = new NesMapper9(mCartridge, mCpu, mPpu);
			break;
		case 10: // mapper #10 - MMC4
			mMapper = new NesMapper10(mCartridge, mCpu, mPpu);
			break;
		case 11: // mapper #11 - Color Dreams
			mMapper = new NesMapper11(mCartridge, mCpu, mPpu);
			break;
		case 16: // mapper #16 - Bandai
			mMapper = new NesMapper16(mCartridge, mCpu, mPpu);
			break;
		case 18: // mapper #18 - Jaleco SS8806
			mMapper = new NesMapper18(mCartridge, mCpu, mPpu);
			break;
		case 19: // mapper #19 - Namcot 106
			mMapper = new NesMapper19(mCartridge, mCpu, mPpu);
			break;
		case 21: // mapper #21 - Konami VRC4
			mMapper = new NesMapper21(mCartridge, mCpu, mPpu);
			break;
		case 22: // mapper #22 - Konami VRC2 type A
			mMapper = new NesMapper22(mCartridge, mCpu, mPpu);
			break;
		case 23: // mapper #23 - Konami VRC2 type B
			mMapper = new NesMapper23(mCartridge, mCpu, mPpu);
			break;
		case 24: // mapper #24 - Konami VRC6
			mMapper = new NesMapper24(mCartridge, mCpu, mPpu);
			break;
		case 25: // Konami VRC4 slight variation of mapper #21
			mMapper = new NesMapper25(mCartridge, mCpu, mPpu);
			break;
		case 32: // mapper #32 - Irem G-101
			mMapper = new NesMapper32(mCartridge, mCpu, mPpu);
			break;
		case 33: // mapper #33 - Taito TC0190
			mMapper = new NesMapper33(mCartridge, mCpu, mPpu);
			break;
		case 34: // mapper #33 - Nina-1
			mMapper = new NesMapper34(mCartridge, mCpu, mPpu);
			break;
		case 64: // mapper #64 - Tengen RAMBO-1
			mMapper = new NesMapper64(mCartridge, mCpu, mPpu);
			break;
		case 65: // mapper #65 - Irem H-3001
			mMapper = new NesMapper65(mCartridge, mCpu, mPpu);
			break;
		case 66: // mapper #66 - GNROM
			mMapper = new NesMapper66(mCartridge, mCpu, mPpu);
			break;
		case 69: // mapper #69 - Sunsoft FME-7
			mMapper = new NesMapper69(mCartridge, mCpu, mPpu);
			break;
		case 71: // mapper #71 - Camerica
			mMapper = new NesMapper71(mCartridge, mCpu, mPpu);
			break;			
		case 75: // mapper #75 - Jaleco/Konami VRC1
			mMapper = new NesMapper75(mCartridge, mCpu, mPpu);
			break;			
		case 78: // mapper #78 - Irem 74HC161/32
			mMapper = new NesMapper78(mCartridge, mCpu, mPpu);
			break;
		case 79: // mapper #79 - Nina-3 AVE
			mMapper = new NesMapper79(mCartridge, mCpu, mPpu);
			break;
		case 228: // mapper #228 - Action 52
			mMapper = new NesMapper228(mCartridge, mCpu, mPpu);
			break;
		case 232: // mapper #232 - Quattro Games
			mMapper = new NesMapper232(mCartridge, mCpu, mPpu);
			break;
		default:
			mMapper = new NesBaseMapper(mCartridge, mCpu, mPpu);
			break;
		}
		
		// Transfer proper data from cartridge to CPU/PPU memory at boot-up depending on mapper
		mMapper->initMemory();
		
		// Set flag that indicates state of NES
		mblnCartLoaded = true;
	}
	
	// Link the Mapper to the CPU
	mCpu->linkCPU(mPpu, mApu, mController1, mController2, mMapper);

	// Call a "finishing" routine, this will be overridden by derived NES classes to provide
	// more functionality during the load (loading SRAM for example)
	finishCartridgeLoad(filename);	

	// Trigger RESET interrupt
	mCpu->mintInterruptLine |= RESET;
	//mCpu->handleInterruptRequest(); // TODO: should this be called here?  or by client? probably here

	// Return status of load
	return mCartridge->mintErrorCode;
}




// Provide a procedure for advancing frame.  This way, a derived NES class can do special things
void Nes::calcFrame() {	
	mCpu->calcFrame();		
}



//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the NES to the file
// This basically involves calling saveState on each component of the NES
//-------------------------------------------------------------------
void Nes::saveState(char *file) {
	FILE *s = fopen(file, "wb");
	if (s != NULL) {
		
		mCpu->saveState(s);
		mPpu->saveState(s);
		mPpu->VRAM->saveState(s);
		mApu->saveState(s);
		mMapper->saveState(s);
		mController1->saveState(s);
		mController2->saveState(s);

		fclose(s);
	}	
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the NES CPU from the file
// This basically involves calling loadState on each component of the NES
//-------------------------------------------------------------------
int Nes::loadState(char *file) {
	FILE *s = fopen(file, "rb");
	if (s == NULL) return 1;
		
	mCpu->loadState(s);
	mPpu->loadState(s);
	mPpu->VRAM->loadState(s);
	mApu->loadState(s);
	mMapper->loadState(s);
	mController1->loadState(s);
	mController2->loadState(s);

	fclose(s);
	
	return 0;		
}